# demo file that shows how to use QauntTree to detect changes in Gaussian data

# A training set of Guassian data N(mu_0, Sigma_0) is generated, where mu_0 and Sigma_0 are random parameters
# A uniform histogram is fitted on the training data using QuantTree
# The, two windows W_0 and W_1 are generated: W_0 is drawn from the same Gaussian used to generate the training set,
#  while W_1 is draw from N(mu_1, Sigma_1), where mu_1 and Sigma_1 are set using CCM to ensure that the
# symmetric kullback-leibler divergence between the two Gaussians distribution is equal to 1
# Four hypothesis testing are then performed both on W_0 and W_1:
# - Pearson with threshold estimated using QuantTree
# - Pearson with threshold set using asymptotic approximation
# - Total Variation with threshold set using QuantTree
# - Total Variation with threshold set using bootstrap

import numpy as np
import libquanttree as qt
import libccm as ccm
from scipy import stats as st

# data dimension
dim = 8

# number of bins of the histogram
K = 32

# number of samples in the training set
N = 4096

# number of sample in the test windws
nu = 64

# target value of the symmetric kullback-leibler divergence between the stationary and alternative distribution
target_sKL = 1

# number of bootstrap resampling
B = 100

# number of sample in the dataset to be bootstrapped
N_boot = 10000

# signifance of the test
alpha = 0.05

# rotate data to align them to coordinate axes
do_PCA = True

# names of considered hypothesis testing
test_names = ['pearson_dist_free', 'pearson_asym', 'tv_dist_free', 'tv_bootstrap']

# generate a random gaussian
gauss0 = ccm.random_gaussian(dim)

# generate random roto-translation used to define the alternative distribution
rot, shift = ccm.compute_roto_translation(gauss0, target_sKL)

# compute the alternative distribution
gauss1 = ccm.rotate_and_shift_gaussian(gauss0, rot, shift)

# generate a stationary dataset
data = np.random.multivariate_normal(gauss0[0], gauss0[1], N)
data_bootstrap = np.random.multivariate_normal(gauss0[0], gauss0[1], N_boot)

# compute the test windows
W0 = np.random.multivariate_normal(gauss0[0], gauss0[1], nu)
W1 = np.random.multivariate_normal(gauss1[0], gauss1[1], nu)

# construct the uniform histogram using QuantTree for each method
qtree = qt.QuantTree(K)
qtree.build_histogram(data, do_PCA)

# construct the change detection tests and set the threshold
all_tests = {}

all_tests['pearson_dist_free'] = qt.ChangeDetectionTest(qtree, nu, qt.pearson_statistic)


threshold = qt.ChangeDetectionTest.get_precomputed_quanttree_threshold('pearson', alpha, K, N, nu)
if threshold is None:
    threshold = all_tests['pearson_dist_free'].estimate_quanttree_threshold(alpha)
all_tests['pearson_dist_free'].set_threshold(alpha, threshold)


all_tests['tv_dist_free'] = qt.ChangeDetectionTest(qtree, nu, qt.tv_statistic)
threshold = qt.ChangeDetectionTest.get_precomputed_quanttree_threshold('tv', alpha, K, N, nu)
if threshold is None:
    threshold = all_tests['tv_dist_free'].estimate_quanttree_threshold(alpha)
all_tests['tv_dist_free'].set_threshold(alpha, threshold)

all_tests['pearson_asym'] = qt.ChangeDetectionTest(qtree, nu, qt.pearson_statistic)
threshold = st.chi2.isf(alpha, K - 1)
all_tests['pearson_asym'].set_threshold(alpha, threshold)

all_tests['tv_bootstrap'] = qt.ChangeDetectionTest(qtree, nu, qt.tv_statistic)
threshold = all_tests['tv_bootstrap'].estimate_bootstrap_threshold(data_bootstrap, alpha, B)
all_tests['tv_bootstrap'].set_threshold(alpha, threshold)


# perform the tests on the two windows W0 and W1
all_reject_null = {}

for test in test_names:
    hp, _ = all_tests[test].reject_null_hypothesis(W0, alpha)
    all_reject_null[(test, 'W0')] = hp

    hp, _ = all_tests[test].reject_null_hypothesis(W1, alpha)
    all_reject_null[(test, 'W1')] = hp
















